"""
This module contains the actual Translator class

BaseTranslator: The super class for all translator classes
Translator: The actual class for doing text translation.
BufferedTranslator: A varient of Translator which performs faster for long text.

"""

import logging
import array
try:
    from psyco.classes import *
except ImportError:
    class psyobj(object):
        pass

from YABT.exceptions import *

logging.getLogger("YABT.translators").manager.emittedNoHandlerWarning = True

class BaseTranslator(psyobj):
    """
    The super class for all translator classes

    Methods are:
    addCharacterMap: Add a character mapping
    addState: Add a state to the known list of states
    addRule: Add a rule to the translator
    load: Load a loader object
    noRuleHandler: A handler method called when a group exists but no rule
    matches.
    getStates: Get a list of states which exist in the translator
    translate: Perform a translation

    """

    def addCharacterMap(self, orig, repl):
        """
        Add a character mapping to the translator

        Arguments are:
        orig: the original character
        repl: the replacement character

        """
        raise NotImplementedError("Class is abstract")

    def addState(self, state):
        """
        Adds a state to the list of known states

        Arguments are:
        state: the state

        """
        raise NotImplementedError("Abstract class")

    def addRule(self, rule):
        """
        Add a rule to the translator

        Arguments are:
        rule: The rule or rule set to add

        """
        raise NotImplementedError("Abstract class")

    def noRuleHandler(self, state, text, start, end):
        """
        Method called when no rule matches

        Arguments are:
        state: The state of the translator
        text: The text being translated
        start: The index of where the focus should begin
        end: Where the focus must end by.

        """
        raise NotImplementedError("Abstract method should be overridden")

    def load(self, loader):
        """
        Load the translation table from the loader

        Arguments are:
        loader: The loader to load the translation table

        """
        raise NotImplementedError("Abstract class")

    def getStates(self):
        """Return a list of the states in this translator"""
        raise NotImplementedError("Abstract class")

    def translate(self, inText, state, bfcontext, afcontext):
        """
        Perform the translation

        Arguments are:
        inText: the text to be translated
        state: The state to start the translator in
        bfcontext: The text which would appear before inText in the source
        afcontext: The text which would appear after inText in the source

        """
        raise NotImplementedError("Abstract class")

class Translator(BaseTranslator):
    """
    The class for doing text translation

    addRule: A method to add a rule to the translator.
    addState: adds a state to the list of known states
    addCharacterMap: Add a character mapping to the translator.
    getStates: Show all valid states for the translator.
    load: Load translation table from loader
    translate: perform translation

    """

    def __init__(self):
        """Create a new YABT translator object"""
        self._logger = logging.getLogger("YABT.translators.Translator")
        if self._logger.level == 0:
            self._logger.setLevel(logging.WARNING)
        self._charMaps = {}
        self._rules = []
        self.state = 0
        self._states = set()

    def addRule(self, rule):
        """
        A method to add a rule to the translator

        Arguments are:
        rule: The rule or rule set object to add

        """
        if self._logger.isEnabledFor(logging.INFO):
            self._logger.info("Adding rule " + str(rule) + " to translator.")
        self._rules.append(rule)

    def addState(self, state):
        """
        Adds a state to the known states

        Arguments are:
        state: The state to add

        """
        if self._logger.isEnabledFor(logging.INFO):
            self._logger.info("Adding state " + str(state))
        self._states.add(state)

    def addCharacterMap(self, orig, repl):
        """
        Adds a character mapping to the translator

        Arguments are:
        orig: The original character.
        repl: the character to replace orig with.

        """
        if self._logger.isEnabledFor(logging.INFO):
            self._logger.info("Adding charmap from " + orig + " to " + repl)
        self._charMaps.update({orig: repl})

    def noRuleHandler(self, state, text, start, end):
        """
        This implementation simply inserts the character into the output
        without altering the state

        """
        return (text[start], 1, state)

    def load(self, loader):
        """
        Loads a translation table from a loader

        Arguments are:
        loader: The loader object to load the translation table from

        """
        loader.load(self)

    def translate(self, text, state, bfcontext=" ", afcontext=" "):
        """
        Performs the actual translation

        Arguments are:
        text: The text to translate
        state: The state to start the translation in.
        bfcontext: The text which is before the translation text, this is not
        actually translated.
        afcontext: The text which is after the translation text, this is not
        actually translated.

        """
        if not state in self.getStates():
            raise InvalidStateException("State " + str(state) + " not found")
        output = array.array("u")
        position = len(bfcontext)
        end = position + len(text)
        fullText = u"".join([bfcontext, text, afcontext])
        for orig, repl in self._charMaps.iteritems():
            fullText = fullText.replace(orig, repl)
        while position < end:
            selectedRule = None
            for rule in self._rules:
                selectedRule = rule.isMatch(state, fullText, position, end)
                if not selectedRule is None:
                    break
            else:
                noRuleInfo = self.noRuleHandler(state, fullText, position, end)
                output.fromunicode(noRuleInfo[0])
                position += noRuleInfo[1]
                if noRuleInfo[2] != -1:
                    state = noRuleInfo[2]
                continue
            output.fromunicode(selectedRule.translation)
            if selectedRule.finalState != -1:
                state = selectedRule.finalState
            position += len(selectedRule.focus)
        self.state = state
        return output.tounicode()

    def getStates(self):
        """Get a set of the states valid for this translator"""
        return self._states

class BufferedTranslator(Translator):
    """
    A performance enhanced version of Translator

    This class improves on performance by splitting the input text into
    lines and processing each line independently. While this improves
    performance, you should note the following:
    As each line is processed separately, no rule which spans more than one
    line can be applied
    Whilst this generally improves performance in some cases it may
    cause a negative impact on performance, eg. those cases where only very short
    blocks of text are being translated

    This class implements the same interface as the Translator class

    """

    def translate(self, text, state, bfcontext=' ', afcontext=' '):
        """
        This performs the translation

        This method performs translation on the input text passed in. This method
        takes the same arguments as the translate method of the Translator
        class. This method probably should perform better than the one in
        Translator, but please make note of the limitations mentioned in the
        class doc string.

        """
        self.state = state
        output = array.array('u')
        inputList = text.splitlines(True)
        inputList.insert(0, bfcontext)
        inputList.append(afcontext)
        beforeContext = inputList[0]
        translateText = inputList[1]
        for afterContext in inputList[2:]:
            output.fromunicode(super(BufferedTranslator,
                                    self).translate(translateText,
                                                    self.state, beforeContext,
                                                    afterContext))
            beforeContext = translateText
            translateText = afterContext
        return output.tounicode()



"""Tests for the translator module"""

import unittest
import os.path

from YABT import loaders, translators
from YABT.exceptions import *

class Translator_SuperTestCase(unittest.TestCase):
    def setUp(self):
        self.translator = translators.Translator()
        self.loader = loaders.BasicXMLStringLoader()
        f = open(os.path.join(os.path.dirname(__file__), "correct.xml"))
        s = f.read()
        f.close()
        self.loader.setInput(s)
        self.translator.load(self.loader)
    def tearDown(self):
        self.loader = None
        self.translator = None

class Translator_BasicTranslate(Translator_SuperTestCase):
    def testTranslate_grade1(self):
        expected = u'THIS SHOULD BE GRADE#A'
        actual = self.translator.translate("this should be grade1", 1, " ", " ")
        self.assertEquals(expected, actual)
    def testTranslate_grade2(self):
        expected = u'? %D 2 GRADE#B'
        actual = self.translator.translate("this should be grade2", 2, " ", " ")
        self.assertEquals(expected, actual)
    def testTranslate_invalidState(self):
        self.assertRaises(InvalidStateException, self.translator.translate, "hello", 0, " ", " ")
    def testGetStates(self):
        expected = set([1,2,3,4,5])
        actual = self.translator.getStates()
        self.assertEquals(expected, actual)

class BufferedTranslator_SuperTestCase(unittest.TestCase):
    def setUp(self):
        self.translator = translators.BufferedTranslator()
        self.loader = loaders.BasicXMLStringLoader()
        f = open(os.path.join(os.path.dirname(__file__), "correct.xml"))
        s = f.read()
        f.close()
        self.loader.setInput(s)
        self.translator.load(self.loader)
    def tearDown(self):
        self.loader = None
        self.translator = None

class BufferedTranslator_BasicTranslate(BufferedTranslator_SuperTestCase):
    def testTranslate_grade1(self):
        expected = "THIS SHOULD BE GRADE#A"
        actual = self.translator.translate("this should be grade1", 1, " ", " ")
        self.assertEquals(expected, actual)
    def testTranslate_grade2(self):
        expected = "? %D 2 GRADE#B"
        actual = self.translator.translate("this should be grade2", 2, " ", " ")
        self.assertEquals(expected, actual)
    def testGetStates(self):
        expected = set([1,2,3,4,5])
        actual = self.translator.getStates()
        self.assertEquals(expected, actual)
    def testTranslate_invalidState(self):
        self.assertRaises(InvalidStateException, self.translator.translate, "hello", 0, " ", " ")




"""
YABT.loaders: a module containing various YABT loader classes

BaseLoader: The class that all loaders should inherit from.
BasicXMLFileLoader: A loader which can load XML files.
BasicXMLStringLoader: A loader to load from a XML string.

"""

from xml.parsers import expat
try:
    from psyco.classes import *
except ImportError:
    class psyobj(object):
        pass

from YABT.basicContextTypes import ContextFactory
from YABT.exceptions import InvalidTranslationTableException
from YABT.rules import GroupStateRuleSet, GeneralContextRule, WordDictRule

class BaseLoader(psyobj):
    """
    This is the abstract class which all loaders should inherit from.
    
    setInput: Set the input source this class should load the table from.
    getInput: Return the input to be used for loading the information from.
    load: load the data into the translator.
    
    """

    def __init__(self):
        """Do not use this as BaseLoader is abstract"""
        super(BaseLoader, self).__init__()
        self._translator = None
        self._input = None

    def setInput(self, inputSource):
        """Set the input source to load the translation table from"""
        self._input = inputSource
    
    def getInput(self):
        """Return the input source for loading the information from"""
        return self._input

    def load(self, translator):
        """Load the translation information into the translator"""
        raise NotImplementedError("Use an actual implementation class")


class BasicXMLFileLoader(BaseLoader):
    """
    A loader to load basic XML translation tables from file

    All methods are as for BaseLoader, except for the below:
    setInput: This takes a file object of a XML file
    load: Overrides load from BaseLoader with an actual implementation
    _startElement: This method is called when the XML parser encounters
    a start tag. This method should only be used by subclasses which override this
    method and require the functionality.
    _endElement: This method is called when the XML parser encounters
    a end tag. This should only be used by subclasses.
    _characterData: This method is called when the XML parser encounters a 
    text element. This should be used only by subclasses.

    """

    def __init__(self, inputsrc=None):
        """Initialises a BasicXMLFileLoader"""
        BaseLoader.__init__(self)
        self.setInput(inputsrc)
        self.contextFactory = ContextFactory()
        self.customMatches = {}
        self._path = ""
        self._iclass = None
        self._trans = None
        self._bfcontext = None
        self._afcontext = None
        self._focus = None
        self._fstate = None
        self._tmpMatch = None
        self._tmpMatchName = None
        self._orig = None
        self._repl = None
        self.__decisions = {}
        self._rules = GroupStateRuleSet()

    def load(self, translator):
        """Load the XML file which has been set into the translator"""
        self._translator = translator
        self.customMatches = {}
        parser = expat.ParserCreate()
        parser.StartElementHandler = self._startElement
        parser.EndElementHandler = self._endElement
        parser.CharacterDataHandler = self._characterData
        try:
            parser.ParseFile(self.getInput())
        except expat.ExpatError:
            raise InvalidTranslationTableException("Table is not valid")
        self.getInput().seek(0)
    def _startElement(self, name, attrs):
        """Method called when a start tag is encountered"""
        self._path += "." + name
        if name == "char":
            self._characterData(unichr(int(attrs["ord"])))
        elif self._path == ".transtable.YABTdata.charmap":
            self._orig = None
            self._repl = None
        elif self._path == ".transtable.YABTdata.charmap.orig":
            self._orig = ""
        elif self._path == ".transtable.YABTdata.charmap.repl":
            self._repl = ""
        elif self._path == ".transtable.YABTdata.decision":
            iclass = int(attrs["inputclass"])
            if not iclass in self.__decisions:
                self.__decisions.update({iclass: set()})
            for state in attrs["states"].split(","):
                self.__decisions[iclass].add(int(state))
                self._translator.addState(int(state))
        elif self._path == ".transtable.YABTdata.rule":
            self._iclass = None
            self._focus = None
            self._bfcontext = None
            self._afcontext = None
            self._trans = None
            self._fstate = None
        elif self._path == ".transtable.YABTdata.rule.iclass":
            self._iclass = ""
        elif self._path == ".transtable.YABTdata.rule.focus":
            self._focus = ""
        elif self._path == ".transtable.YABTdata.rule.bfcontext":
            self._bfcontext = ""
        elif self._path == ".transtable.YABTdata.rule.afcontext":
            self._afcontext = ""
        elif self._path == ".transtable.YABTdata.rule.trans":
            self._trans = ""
        elif self._path == ".transtable.YABTdata.rule.fstate":
            self._fstate = ""
        elif self._path.startswith(".transtable.metadata.YABT_custom_matches."):
            self._tmpMatchName = name
            self._tmpMatch = ""

    def _endElement(self, name):
        """Method called when an end tag is encountered"""
        if name == "char":
            pass
        elif self._path == ".transtable.YABTdata.charmap":
            self._translator.addCharacterMap(self._orig, self._repl)
        elif self._path == ".transtable.YABTdata.rule":
            self._addRule()
        elif self._path.startswith(".transtable.metadata.YABT_custom_matches."):
            self.customMatches.update({self._tmpMatchName: self._tmpMatch})
        elif self._path == ".transtable":
            self._translator.addRule(self._rules)
        self._path = self._path[:-1 - len(name)]

    def _characterData(self, data):
        """Method called when a text element is encountered"""
        if self._path == ".transtable.YABTdata.charmap.orig":
            self._orig += data
        elif self._path == ".transtable.YABTdata.charmap.repl":
            self._repl += data
        elif self._path == ".transtable.YABTdata.rule.iclass":
            self._iclass += data
        elif self._path == ".transtable.YABTdata.rule.focus":
            self._focus += data
        elif self._path == ".transtable.YABTdata.rule.bfcontext":
            self._bfcontext += data
        elif self._path == ".transtable.YABTdata.rule.afcontext":
            self._afcontext += data
        elif self._path == ".transtable.YABTdata.rule.trans":
            self._trans += data
        elif self._path == ".transtable.YABTdata.rule.fstate":
            self._fstate += data
        elif self._path.startswith(".transtable.metadata.YABT_custom_matches."):
            self._tmpMatch += data

    def _addRule(self):
        """Private method to prepare data and adding rule"""
        iclass = int(self._iclass)
        focus = unicode(self._focus)
        trans = unicode(self._trans)
        fstate = int(self._fstate)
        bfcontext = self._bfcontext[2:]
        afcontext = self._afcontext[2:]
        bftype = self._bfcontext[:2]
        aftype = self._afcontext[:2]
        if bftype == "^c":
            bfcontext = self.customMatches[bfcontext]
        if aftype == "^c":
            afcontext = self.customMatches[afcontext]
        bfContextObj = self.contextFactory.getContextInstance(bftype,
                                bfcontext, self.contextFactory.BEFORE)
        afContextObj = self.contextFactory.getContextInstance(aftype,
                                afcontext, self.contextFactory.AFTER)
        group = focus[0]
        rule = GeneralContextRule(focus, bfContextObj, afContextObj, trans, 
                                  fstate)
        for state in self.__decisions[iclass]:
            self._rules.addRule(rule, (group, state))

class BasicXMLStringLoader(BasicXMLFileLoader):
    """
    A loader to load XML from a string

    All methods are as for BasicXMLFileLoader with the following exception:
    load: This overrides the BasicXMLFileLoader version and so works on 
    a string of XML rather than a file type object

    """

    def __init__(self):
        """Create a XML string loader"""
        super(BasicXMLStringLoader, self).__init__()

    def load(self, translator):
        """Load the XML string set into the translator"""
        self._translator = translator
        self.customMatches = {}
        parser = expat.ParserCreate()
        parser.StartElementHandler = self._startElement
        parser.EndElementHandler = self._endElement
        parser.CharacterDataHandler = self._characterData
        try:
            parser.Parse(self.getInput())
        except expat.ExpatError:
            raise InvalidTranslationTableException("Table is not valid")


class BasicWordDictLoader(BaseLoader):
    """Load a WordDictRule from a simple text file"""
    def __init__(self, inputsrc=None):
        """Create a WordDictLoader with optional input source"""
        super(BasicWordDictLoader, self).__init__()
        self.setInput(inputsrc)

    def load(self, translator):
        """Load the word dictionary into the translator"""
        input = self.getInput()
        originalPosition = input.tell()
        wd = WordDictRule()
        for line in input.readlines():
            if len(line) == 0 or not line[0].isdigit():
                continue
            s = line.split()
            if len(s) != 3:
                continue
            state = 0
            try:
                state = int(s[0])
            except ValueError:
                continue
            wd.addWord(state, s[1], s[2])
        translator.addRule(wd)
        input.seek(originalPosition)




"""
This module contains the basic context types for YABT

Classes:
ContextFactory: This is a factory class for context objects
ContextSuper: This class should be the super class to all contextObjects
All other classes are only for internal use and only need be known to other
classes as implementations of ContextSuper.

"""

import re

try:
    from psyco.classes import *
except ImportError:
    class psyobj(object):
        pass

class ContextSuper(psyobj):
    """
    This is the super class that all context objects should implement

    This class is an abstract class which all context objects should implement.
    Subclasses should override the following:
    __init__(self, pattern): This constructor is the one called by the factory
    class. The pattern argument is the string used to specify how this context
    object should match.
    match(self, text, start, end): This method is called to check whether the
    context matches. The text argument is the complete string, start is the
    start index of the focus and end is the index of the end of the focus
    (IE. text[start:end] should give the focus).

    """
    
    def match(self, text, start, end):
        """Abstract method to check whether the context matches"""
        raise NotImplementedError

    def __init__(self, pattern):
        """Create a context object with the given pattern"""
        self.pattern = pattern
        self.priority = 0

class ContextA(ContextSuper):
    """This is a context type for all contexts to match"""
    def match(self, text, start, end):
        """Returns True regardless of the actual context"""
        return True

class ContextCSuper(ContextSuper):
    """This is the super class for custom match contexts"""
    def __init__(self, pattern):
        """This class is abstract, this merely partially sets up"""
        ContextSuper.__init__(self, pattern)
        self.priority = 2
    def match(self, text, start, end):
        """Reports whether the context matches"""
        raise NotImplementedError
class ContextCBefore(ContextCSuper):
    """Checks if the character before is in the list of characters"""
    def match(self, text, start, end):
        """Only matches if the adjacent character is in the custom list"""
        return text[start - 1] in self.pattern

class ContextCAfter(ContextCSuper):
    def match(self, text, start, end):
        return text[end] in self.pattern

class ContextTSuper(ContextSuper):
    def __init__(self, pattern):
        ContextSuper.__init__(self, pattern)
        self.priority = 3

class ContextTBefore(ContextTSuper):
    def match(self, text, start, end):
        return text.endswith(self.pattern, 0, start)

class ContextTAfter(ContextTSuper):
    def match(self, text, start, end):
        return text.startswith(self.pattern, end)

class ContextRBefore(ContextSuper):
    def __init__(self, pattern):
        ContextSuper.__init__(self, pattern + "$")
        self._compiled = re.compile(self.pattern)
        self.priority = 1
    def match(self, text, start, end):
        if self._compiled.search(text, 0, start):
            return True
        else:
            return False

class ContextRAfter(ContextSuper):
    def __init__(self, pattern):
        ContextSuper.__init__(self, "^" + pattern)
        self._compiled = re.compile(self.pattern)
        self.priority = 1
    def match(self, text, start, end):
        if self._compiled.search(text, end):
            return True
        else:
            return False

class ContextFactory:
    """
    This class creates the requested context object

    Methods are:
    getContextInstance: this is the method called to get a context object.
    addContextType: Adds a context type to the factory.

    """

    BEFORE = 0
    AFTER = 1
    def __init__(self):
        """Create a ContextFactory for getting context objects"""
        self._contextObjects = {}
        self._contextTypes = {"^a": (ContextA, ContextA),
                               "^c": (ContextCBefore, ContextCAfter),
                               "^t": (ContextTBefore, ContextTAfter),
                               "^r": (ContextRBefore, ContextRAfter)}

    def addContextType(self, name, before, after):
        """
        Add a new context type to the context factory

        Arguments are:
        name: The string used to identify the type
        before: The class to be used for before contexts
        after: The class to be used for after contexts

        """
        self._contextTypes.update({name: (before, after)})

    def getContextInstance(self, ctType, pattern, position):
        """
        This method will create a context object according to arguments

        Arguments are:
        type: The string to identify the type to be created
        pattern: The string which defines how the context should match
        position: This should be one of ContextFactory.BEFORE or
        ContextFactory.AFTER

        """
        try:
            obj = self._contextObjects[(ctType, pattern, position)]
        except KeyError:
            obj = self._contextTypes[ctType][position](pattern)
            self._contextObjects.update({(ctType, pattern, position): obj})
        return obj



"""
This module provides the classes for rules

BaseRule: An abstract class which all rule like classes should inherit from
RuleSet: A basic class for rule sets, not intended to be used directly.
GeneralContextRule: A rule to match nearly any context
GroupStateRuleSet: This groups its rules by group and state

"""

try:
    from psyco.classes import psyobj
except ImportError:
    class psyobj(object):
        def __init__(self):
            """Initialize psyobj stand in class"""
            object.__init__(self)

class BaseRule(psyobj):
    """Abstract class for all rule like objects to inherit"""

    def __init__(self):
        """Should not be used to create an object, only for subclasses"""
        super(BaseRule, self).__init__()
    def isMatch(self, state, text, position, end):
        """
        Determines whether this rule can match. If this rule matches it
        returns the rule object otherwise returns None.
        """
        raise NotImplementedError("Abstract class")

class RuleSet(BaseRule):
    """
    This is a basic rule set, it will check whether any of its rules matches,
    it does no condition checking itself, so is mainly useful fo a super class to more advanced rule sets

    """
    def __init__(self):
        """Create a basic rule set"""
        BaseRule.__init__(self)
        self.__ruleSet = []

    def findRule(self, state, text, position, end):
        """
        This finds the first matching rule, it is intended to be called from
        this class or a subclass

        """
        for rule in self.__ruleSet:
            matchingRule = rule.isMatch(state, text, position, end)
            if not matchingRule is None:
                return matchingRule
        return None

    def addRule(self, rule):
        """Add a rule to the rule set"""
        self.__ruleSet.append(rule)

    def isMatch(self, state, text, position, end):
        """
        Finds the first matching rule of the set and returns it. If no
        match is found then None is returned.

        """
        return self.findRule(state, text, position, end)

class GeneralContextRule(BaseRule):
    """
    Rule to check whether the current position matches the focus and that the
    context matches. Nearly any context should be possible to specify.

    """
    def __init__(self, focus, bfcontext, afcontext, trans, fstate):
        """Create a general context rule"""
        super(GeneralContextRule, self).__init__()
        self.focus = focus
        self.translation = trans
        self.finalState = fstate
        if bfcontext.priority < afcontext.priority:
            self.__context = (afcontext, bfcontext)
        else:
            self.__context = (bfcontext, afcontext)

    def isMatch(self, state, text, position, end):
        """
        Checks whether this rule actually matches. If it matches it returns
        itself otherwise returns None

        """
        if text.startswith(self.focus, position, end):
            focusEnd = position + len(self.focus)
            if self.__context[0].match(text, position, focusEnd):
                if self.__context[1].match(text, position, focusEnd):
                    return self
        return None

class GroupStateRuleSet(psyobj):
    """This rule set groups rules according to group and state all in one"""
    def __init__(self):
        """Create a group and state rule set"""
        super(GroupStateRuleSet, self).__init__()
        self.__groups = {}
    def addRule(self, rule, group):
        """
        Add a rule to the group

        Arguments are:
        rule: The rule to add
        group: this is the group and state the rule belongs to in a tuple

        """
        if not group in self.__groups:
            self.__groups.update({group: []})
        self.__groups[group].append(rule)

    def isMatch(self, state, text, position, end):
        """Find the rule or return None if no rule matches"""
        group = text[position]
        if not (group, state) in self.__groups:
            return None
        for rule in self.__groups[(group, state)]:
            selectedRule = rule.isMatch(state, text, position, end)
            if not selectedRule is None:
                return selectedRule
        return None

class WordDictRule(BaseRule):
    """This class provides a rule to find words from a dictionary"""

    def __init__(self):
        """Create a blank word dictionary rule"""
        super(WordDictRule, self).__init__()
        self._words = {}

    def isMatch(self, state, text, position, end):
        """Checks for a matching word"""
        self.focus = None
        self.translation = None
        self.finalState = state
        # Check that we are at the beginning of a word
        if text[position - 1].isalpha():
            return None
        # Find the length of the word at position
        index = position
        word = ""
        while index < end:
            if not text[index].isalpha():
                word = text[position:index]
                break
            index += 1
        try:
            self.translation = self._words[state][word]
        except KeyError:
            return None
        else:
            self.focus = word
            return self
        return None

    def addWord(self, state, orig, repl):
        """Add a word to the dictionary"""
        if not state in self._words:
            self._words.update({state: {}})
        self._words[state].update({orig: unicode(repl)})




"""This module tests YABT.loaders"""

import unittest
import os.path

from YABT import loaders
from YABT.exceptions import InvalidTranslationTableException

class FakeTranslator:
    def __init__(self):
        self.charMaps = 0
        self.states = 0
        self.rules = 0
    def addCharacterMap(self, orig, repl):
        self.charMaps += 1
    def addState(self, state):
        self.states += 1
    def addRule(self, rule):
        self.rules += 1
    def load(self, loader):
        loader.load(self)

class BasicXMLFileLoader_SuperTestCase(unittest.TestCase):
    def setUp(self):
        self.loader = loaders.BasicXMLFileLoader()
    def tearDown(self):
        self.loader = None

class BasicXMLStringLoader_SuperTestCase(unittest.TestCase):
    def setUp(self):
        self.loader = loaders.BasicXMLStringLoader()
    def tearDown(self):
        self.loader = None

class BasicXMLFileLoader_ValidRun(BasicXMLFileLoader_SuperTestCase):
    def setUp(self):
        BasicXMLFileLoader_SuperTestCase.setUp(self)
        f = open(os.path.join(os.path.dirname(__file__), "correct.xml"))
        self.loader.setInput(f)
    def tearDown(self):
        self.loader.getInput().close()
        BasicXMLFileLoader_SuperTestCase.tearDown(self)
    def testLoad(self):
        t = FakeTranslator()
        t.load(self.loader)
        self.assertEquals(28, t.charMaps)
        self.assertEquals(13, t.states)
        self.assertEquals(1, t.rules)

class BasicXMLFileLoader_BadlyFormedTable(BasicXMLFileLoader_SuperTestCase):
    def setUp(self):
        BasicXMLFileLoader_SuperTestCase.setUp(self)
        f = open(os.path.join(os.path.dirname(__file__), "badlyformed.xml"))
        self.loader.setInput(f)
    def tearDown(self):
        self.loader.getInput().close()
        BasicXMLFileLoader_SuperTestCase.tearDown(self)
    def testLoad(self):
        t = FakeTranslator()
        self.assertRaises(InvalidTranslationTableException, t.load, self.loader)

class BasicXMLStringLoader_ValidRun(BasicXMLStringLoader_SuperTestCase):
    def setUp(self):
        BasicXMLStringLoader_SuperTestCase.setUp(self)
        f = open(os.path.join(os.path.dirname(__file__), "correct.xml"))
        s = f.read()
        f.close()
        self.loader.setInput(s)
    def testLoad(self):
        t = FakeTranslator()
        t.load(self.loader)
        self.assertEquals(28, t.charMaps)
        self.assertEquals(13, t.states)
        self.assertEquals(1, t.rules)

class BasicXMLStringLoader_BadlyFormed(BasicXMLStringLoader_SuperTestCase):
    def setUp(self):
        BasicXMLStringLoader_SuperTestCase.setUp(self)
        f = open(os.path.join(os.path.dirname(__file__), "badlyformed.xml"))
        s = f.read()
        f.close()
        self.loader.setInput(s)
    def testLoad(self):
        t = FakeTranslator()
        self.assertRaises(InvalidTranslationTableException, t.load, self.loader)




"""
This module provides the exception classes for YABT

InvalidStateException: Exception to indicate that the state does not exist
InvalidTranslationTableException: Indicates that the table does not
conform to the format.

"""

class InvalidTranslationTableException(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class InvalidStateException(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)




#!/usr/bin/python
import ez_setup
ez_setup.use_setuptools()
from setuptools import setup
import logging
__version = "Unknown"
try:
    import version
except ImportError:
    f = open("version.txt")
    __version__ = f.read()
    f.close()
else:
    __version__ = version.getVersion("2.0.1Update1")
    f = open("version.txt", 'w')
    f.write(__version__)
    f.close()
log = logging.root
setup(name='YABT',
version=__version__,
author='Michael Whapples',
author_email='mwhapples@users.berlios.de',
url='http://yabt.berlios.de',
description='YABT: A context sensitive string manipulation tool',
long_description='''
YABT: A context sensitive string manipulation tool

YABT is a context sensitive string manipulation tool written in pure python. It has two main goals:
* Provide implementations of string manipulation for specific tasks to be used inside applications. The current 
implementation is a context matching rule system combined with a finite state machine and a table for translating text 
into British Braille. This implementation should be useful for other tasks such as preparing text for a speech 
synthesiser by providing different tables.
* To provide a framework which developers can use to create new translation algorithms which can be used by any 
application using YABT. This side of YABT could do with further development and so is most likely going t see some 
changes as YABT matures.
''',
license="RPL",
classifiers=['Development Status :: 5 - Production/Stable',
	'Intended Audience :: Developers',
	'License :: OSI Approved',
	'Operating System :: OS Independent',
	'Programming Language :: Python',
	'Topic :: Adaptive Technologies',
	'Topic :: Software Development'
],
packages=['YABT'],
package_data={'YABT': ['tables/britishtobrl.xml']},
entry_points={'console_scripts':['yabt-trans = YABT.app:standardTranslate']},
test_suite="tests"
)




"""A module containing functions for scripts"""

import sys
import optparse
import os.path
from pkg_resources import *

from YABT import loaders, translators

distro = get_distribution("YABT")

def standardTranslate():
    """YABT_translate script"""
    global distro
    inFile = None
    parser = optparse.OptionParser(usage="%prog [OPTIONS] INFILE",
                                   version="%prog " + distro.version)
    parser.add_option("-t", "--table", dest="table", 
                      help="use the table specified for translation " +
                           "[DEFAULT: %default]")
    parser.add_option("-s", "--state", dest="state", type="int",
                      help="Specify the state to use, usually 1 for grade1" +
                      "or 2 for grade2 [DEFAULT: %default]")
    parser.add_option("-o", "--output", dest="outfile",
                      help="Output to a file instead of STDOUT " +
                      "[DEFAULT %default]")
    parser.add_option("-d", "--dictionary", dest="dictionary",
                      help="use the specified word substitution " +
                      "dictionary as well as the standard rules " +
                      "[default: %default]")
    parser.set_defaults(table="britishtobrl.xml", state=2, outfile=None,
                        dictionary=None)
    options, args = parser.parse_args()
    loader = loaders.BasicXMLStringLoader()
    translator = translators.BufferedTranslator()
    if options.dictionary is not None:
        wdl = loaders.BasicWordDictLoader()
        try:
            wdl.setInput(open(options.dictionary))
        except IOError:
            parser.error("Word dictionary couldn't be opened")
        translator.load(wdl)
        wdl.getInput().close()
    if resource_exists("YABT", "tables/" + options.table):
        loader.setInput(resource_string("YABT", "tables/" + options.table))
    elif os.path.exists(options.table):
        try:
            inFile = open(options.table, 'r')
            loader.setInput(inFile.read())
        except IOError: parser.error("Problem reading table")
        finally: inFile.close()
    else:
        parser.error("Table couldn't be found")
    inText = ""
    if len(args) == 0:
        line = sys.stdin.readline()
        while line != "":
            inText += unicode(line, 'latin1')
            line = sys.stdin.readline()
    elif len(args) > 1:
        parser.error("Only one input file should be specified")
    elif os.path.exists(args[0]):
        try:
            inFile = open(args[0], 'r')
            inText = unicode(inFile.read(), "latin1")
        except IOError: parser.error("INFILE could not be read")
        finally: inFile.close()
    else:
        parser.error("INFILE not found")
    translator.load(loader)
    output = translator.translate(inText, options.state, " ", " ")
    if options.outfile is None:
        sys.stdout.write(output)
    else:
        try:
            f = open(options.outfile, 'w')
            f.write(output)
            f.close()
        except:
            parser.error("Can't write output file")
    return 0

if "__main__" == __name__:
    standardTranslate()
